#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/vt.h>
#include "fb.h"

int main(int argc, char* argv[]) {
	int q;
	int fd;
	int fdt;
	int err;
	struct fb_con2fbmap c2f;
	struct fb_fix_screeninfo fix;
	struct fb_var_screeninfo var;
	char x[100];
	int ok;	
	
	if (argc < 2) {
		fprintf(stderr, "SETVT vtnumber\n");
		return 123;
	}
	q = strtoul(argv[1], NULL, 0);
	sprintf(x, "/dev/tty%u", q);
	printf("Determining status of /dev/tty%u\n", q);
	fdt = open(x, O_RDWR);
	if (fdt == -1) {
		fprintf(stderr, "Something is wrong with %s: %s\n",
			x, strerror(errno));
		close(fdt);
		return 121;
	}
	sprintf(x, "/dev/fb0");
	fd = open(x, O_RDONLY);
	if (fd == -1) {
failopen:;
		if ((errno == ENOENT) || (errno == ENXIO) || (errno == ENODEV))
			fprintf(stderr, "Probably non-fb computer\n");
		else if ((errno == EPERM) || (errno == EACCES))
			fprintf(stderr, "You have not enough rights to control fb (maybe fb not instaled)\n");
		else
			perror("Cannot open /dev/fbX");
		fprintf(stderr, "Giving up\n");
		close(fdt);
		return 123;
	}
	c2f.console = q;
	c2f.framebuffer = ~0;
	err = ioctl(fd, FBIOGET_CON2FBMAP, &c2f);
	if (err) {
		perror("Cannot get CON2FB mapping");
		close(fd);
		fprintf(stderr, "Giving up\n");
		return 122;
	}
	printf("/dev/tty%u belongs to /dev/fb%u\n", q, c2f.framebuffer);
	if (c2f.framebuffer != 0) {
		/* close fb0 and open fbX */
		close(fd);
		sprintf(x, "/dev/fb%u", c2f.framebuffer);
		fd = open(x, O_RDONLY);
		if (fd == -1)
			goto failopen;
	}
	/* WE MUST activate VT before we can set videomode */
	err = ioctl(fdt, VT_ACTIVATE, q);
	if (err) {
		perror("Cannot set FG terminal");
		close(fdt);
		close(fd);
		return 120;
	}
	err = ioctl(fdt, VT_WAITACTIVE, q);
	if (err) {
		perror("Cannot set FG terminal (2)");
		close(fdt);
		close(fd);
		return 120;
	}
	close(fdt);
	/* OK, active... */
	err = ioctl(fd, FBIOGET_FSCREENINFO, &fix);
	if (err) {
		perror("Cannot get current parameters");
		close(fd);
		return 118;
	}
	if (strcmp(fix.id, "MATROX")) {
		close(fd);
		fprintf(stderr, "Unknown graphics engine %s\n", fix.id);
		return 117;
	}
	err = ioctl(fd, FBIOGET_VSCREENINFO, &var);
	if (err) {
		perror("Cannot get current parameters");
		close(fd);
		return 119;
	}
	ok = 1;
	/* bpp = 0 -> switch to VGA mode */
	if (var.bits_per_pixel != 0) {
		var.bits_per_pixel = 0;
		ok = 0;
	}
	/* MATROX does not handle clocks faster than 60MHz in text mode */
	if (var.pixclock < 15000) {
		var.pixclock = 15000;
		ok = 0;
	}
	/* switch to 6bit DAC */
	if ((var.red.length != 6) || (var.green.length != 6) || (var.blue.length != 6)) {
		var.red.length = var.green.length = var.blue.length = 6;
		var.red.offset = var.green.offset = var.blue.offset = 0;
		ok = 0;
	}
	if (!ok) {
		var.activate = FB_ACTIVATE_NOW;
		var.nonstd = 0;	/* to be safe ... */
		err = ioctl(fd, FBIOPUT_VSCREENINFO, &var);
		if (err) {
			perror("Cannot set requested videomode :-(");
			close(fd);
			return 116;
		}
		printf("Parameters set\n");
	} else
		printf("Parameters already set\n");
	close(fd);
	printf("Requested mode was set on console %u\n", q);
	return 0;
}
